<?php defined('SYSPATH') or die('No direct script access.');

class Model_Rating extends Model
{
    public static function GetStudentsForRating($discipline) {
        $sql = 'CALL `GetStudentsForRating`(:discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->execute();
    }

    /**
     * @param $discipline int discipline id
     * @param $student    int student id
     * @return Database_Result
     */
    public static function getRates($discipline, $student) {
        $sql = 'CALL `GetRates`(:student, :discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':student', $student)
            ->execute();
    }

    /**
     * @param $discipline int discipline id
     * @param $student    int student id
     * @return Database_Result
     */
    public static function getRatesAll($discipline, $student) {
        $sql = 'CALL `GetRatesAll`(:student, :discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':student', $student)
            ->execute();
    }


    public static function getMapForStudent($student, $discipline) {
        $sql = 'CALL `GetMapForStudent`(:student, :discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':student', $student)
            ->execute();
    }

    public static function getAttestationData($discipline, $group) {
        $sql = 'CALL `GetAttestationData`(:discipline, :group)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':group', $group)
            ->execute();
    }

    private static function checkRateChanges($studentID, $disciplineID, $submoduleID, $rate) {
        $discipline = Model_Discipline::load($disciplineID);

        $map = Model_Map::of($discipline);
        $modules = $map->getModules('all');
        $submodules = $map->getSubmodules('all');
        foreach ($submodules as &$submodule) {
            $moduleID = $submodule['ModuleID'];
            $submodule['ModuleType'] = $modules[$moduleID]['Type'];
        }

        $rates = Model_Rating::getRatesAll($disciplineID, $studentID);

        // Информация об оценке
        $newRateInfo = [
            SubmoduleID => $submoduleID,
            Rate => $rate != -1 && $rate != 'null' ? $rate : NULL,
        ];
        // Поиск типа оценки, доборов и экзаменов
        $extras = [];
        $exams = [];
        foreach ($submodules as &$submodule) {
            $subID = $submodule['ID'];
            $type = $submodule['ModuleType'];
            if ($newRateInfo['SubmoduleID'] == $subID) $newRateInfo['ModuleType'] = $type;
            if ($type == 'extra') $extras[$subID] = NULL;
            if ($type == 'exam') $exams[$subID] = NULL;
        }
        $isExam = $discipline['Type'] == 'exam';

        // Считаем сумму оценок
        $rating = [
            regular => 0,
            extra   => 0,
        ];
        $tryExtra = false;
        $tryExam = false;
        foreach ($rates as $rateInfo) {
            $subID = $rateInfo['SubmoduleID'];
            $type = $rateInfo['ModuleType'];
            $rate = !is_null($rateInfo['ExamPeriodOption']) ? $rateInfo['ExamPeriodOption'] : $rateInfo['Rate'];
            $newRate = $newRateInfo['SubmoduleID'] == $subID ? $newRateInfo['Rate'] : $rate;

            switch ($type) {
                // Оценка семестра - суммируем
                case 'regular':
                    $rating['regular'] += $newRate;
                    break;

                // Оценка добора - суммируем, обновляем значение в массиве доборов и ставим флаг добора
                case 'extra':
                    $rating['extra'] += $newRate;
                    $extras[$subID] = $newRate;
                    if (!is_null($newRate))
                        $tryExtra = true;
                    break;

                // Оценка экзамена - обновляем значение в массиве экзаменов и ставим флаг экзамена
                case 'exam':
                    $exams[$subID] = $newRate;
                    if (!is_null($newRate))
                        $tryExam = true;
                    break;
            }
        }

        // Переделы баллов в семестре
        $semRates = $isExam ? [
            min => 38,
            max => 60,
        ] : [
            min => 60,
            max => 100,
        ];

        // Работаем в зависимости от типа оценки
        switch ($newRateInfo['ModuleType']) {
            case 'regular':
                if ($tryExam && $rating['regular'] + $rating['extra'] < $semRates['min'])
                    return 'Недостаточно баллов для сдачи экзамена';
                
                if (array_shift($exams) == 'pass' && $rating['regular'] < 60)
                    return 'Недостаточно баллов для получения автомата';

                if ($tryExtra && $rating['regular'] + $rating['extra'] > $semRates['min'])
                    return 'Превышение максимума баллов для добора';

                break;

            case 'extra':
                if ($tryExam && $rating['regular'] + $rating['extra'] < $semRates['min'])
                    return 'Недостаточно баллов для сдачи экзамена';

                if ($rating['regular'] + $rating['extra'] > $semRates['min'])
                    return 'Превышение максимума баллов для добора';


                // Поиск предыдущего и следующего добора
                $extrasList = array_keys($extras);
                $pos = array_search($newRateInfo['SubmoduleID'], $extrasList);
                $firstExtra = $pos == 0;
                $lastExtra = $pos == count($extras) - 1;
                $prevExtra = !$firstExtra ? $extras[$extrasList[$pos - 1]] : NULL;
                $nextExtra = !$lastExtra ? $extras[$extrasList[$pos + 1]] : NULL;

                if (!$firstExtra && is_null($prevExtra))
                    return 'Невозможно выставление баллов добора при отсутствии предыдущего добора';

                if (is_null($newRateInfo['Rate']) && !$lastExtra && !is_null($nextExtra))
                    return 'Невозможно удаление баллов добора при наличии следующего добора';

                break;

            case 'exam':
                if ($rating['regular'] + $rating['extra'] < $semRates['min'])
                    return 'Недостаточно баллов для сдачи экзамена';


                // Поиск предыдущей и следующей попыток
                $examsList = array_keys($exams);
                $pos = array_search($newRateInfo['SubmoduleID'], $examsList);
                $firstExam = $pos == 0;
                $lastExam = $pos == count($exams) - 1;
                $prevExam = !$firstExam ? $exams[$examsList[$pos - 1]] : NULL;
                $nextExam = !$lastExam ? $exams[$examsList[$pos + 1]] : NULL;


                if (!$firstExam && is_null($prevExam))
                    return 'Невозможно выставление баллов попытки сдачи экзамена при отсутствии предыдущей попытки';
                
                if (!$firstExam && ($prevExam >= 22 || $prevExam == 'pass'))
                    return 'Для студента сдавшего экзамен следующие пересдачи невозможны';
                
                if (is_null($newRateInfo['Rate']) && !$lastExam && !is_null($nextExam))
                    return 'Невозможно удаление баллов сдачи при наличии следующей попытки сдачи';
                
                if (($newRateInfo['Rate'] >= 22 || $newRateInfo['Rate'] == 'pass') && !$lastExam && !is_null($nextExam))
                    return 'Попытка сдачи экзамена не может быть успешной при наличии следующей попытки сдачи';

                if ($newRateInfo['Rate'] == 'pass' && $rating['regular'] < 60)
                    return 'Недостаточно баллов для автомата';

                if ($newRateInfo['Rate'] == 'pass' && !$firstExam)
                    return 'Невозможно выставление автомата не первой попытке сдачи';

                break;

            case 'bonus':

                break;
        }
    }

    public static function SetRate($teacherID, $studentID, $disciplineID, $submoduleID, $rate) {
        $errorMsg = Model_Rating::checkRateChanges($studentID, $disciplineID, $submoduleID, $rate);
        if (!is_null($errorMsg)) return $errorMsg;

        $errorMsg = Model_Rating::SetRateSQL($teacherID, $studentID, $submoduleID, $rate);
        if (!is_null($errorMsg)) return $errorMsg;

        $errorMsg = Model_Rating::SetExamPeriodOptionSQL($studentID, $submoduleID, 'null');
        if (!is_null($errorMsg)) return $errorMsg;
    }

    public static function SetExamPeriodOption($teacherID, $studentID, $disciplineID, $submoduleID, $option) {
        $errorMsg = Model_Rating::checkRateChanges($studentID, $disciplineID, $submoduleID, $option);
        if (!is_null($errorMsg)) return $errorMsg;

        $errorMsg = Model_Rating::SetRateSQL($teacherID, $studentID, $submoduleID, '-1');
        if (!is_null($errorMsg)) return $errorMsg;

        $errorMsg = Model_Rating::SetExamPeriodOptionSQL($studentID, $submoduleID, $option);
        if (!is_null($errorMsg)) return $errorMsg;
    }

    public static function SetRateSQL($teacherID, $studentID, $submoduleID, $rate) {
        $sql = 'SELECT `Discipline_SetRate`(:teacher, :student, :submodule, :rate) AS `Num`';
        $dbAnswer = DB::query(Database::SELECT, $sql)
            ->parameters([
                ':teacher'   => $teacherID,
                ':student'   => $studentID,
                ':submodule' => $submoduleID,
                ':rate'      => (int) $rate
            ])->execute();

        $result = $dbAnswer->get('Num');
        if ($result != 0) {
            switch ($result) {
                case 1: return 'Недостаточно прав';
                case 2: return 'Семестр завершен, выставление баллов запрещено';
                case 3: return 'Превышение максимального балла';
                default: HTTP_API_Exception::factory(500, 'Ошибка приложения');
            }
        }
    }

    public static function SetExamPeriodOptionSQL($studentID, $submoduleID, $option) {
        $sql = 'SELECT `Discipline_SetExamPeriodOption`(:student, :submodule, :option) as `Num`';
        $dbAnswer = DB::query(Database::SELECT, $sql)
            ->parameters([
                ':student'   => $studentID,
                ':submodule' => $submoduleID,
                ':option'    => $option
            ])->execute();

        $result = $dbAnswer->get('Num');
        if ($result != 0) {
            HTTP_API_Exception::factory(500, 'Ошибка приложения');
        }
    }

    public static function getRatesForDiscipline(Model_Discipline $discipline) {
        $sql = 'CALL `GetRatesForDiscipline`(:id)';
        $result = DB::query(Database::SELECT, $sql)
            ->param(':id', $discipline->ID)->execute();

        $table = [];
        foreach ($result as $row) {
            $student = $row['StudentID'];
            $submodule = $row['SubmoduleID'];
            $table[$student][$submodule] = $row['Rate'];
        }

        return $table;
    }

    public static function getRatesForStudentsGroup($discipline, $group) {
        $sql = 'CALL `GetRatesForGroup`(:discipline, :group)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':discipline' => $discipline,
                ':group'      => $group,
            ])->execute();
    }

    public static function getRatesForStudentsGroupByStage($discipline, $group, $stage) {
        $sql = 'CALL `GetRatesForGroupByStage`(:discipline, :group, :stage)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':discipline' => $discipline,
                ':group'      => $group,
                ':stage'      => $stage
            ])->execute();
    }

    /** @return  Database_RESULT   Database_Result for SELECT queries */
    public static function getRatesForGroupAll($group, $semester) {
        $sql = 'CALL `GetRatesForGroupAll`(:group, :semester)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':group'    => $group,
                ':semester' => $semester
            ])->execute();
    }

    public static function getFinalFormInfo($discipline, $group) {
        $sql = 'CALL `getFinalFormInfo`(:discipline, :group)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':discipline' => $discipline,
                ':group'      => $group,
            ])->execute()[0];
    }

    /**
     * Get count of related with discipline records in `rating_table`.
     * @param Model_Discipline $discipline
     * @return int
     */
    public static function count(Model_Discipline $discipline) {
        $sql = 'SELECT `Discipline_CountRatings`(:discipline) AS `res`';
        $res = DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline->ID)
            ->execute();

        return (int) $res->get('res');
    }

    public static function getHistory($discipline) {
        $sql = 'CALL `GetRatesHistory`(:discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->execute();
    }
};